
from pathlib import Path
from typing import Dict

DEFAULT_HOT_STORAGE_LIMIT = 1 << 30  # 1 GiB
DEFAULT_COLD_STORAGE_LIMIT = 10 << 30  # 10 GiB
DEFAULT_COLD_STORAGE_ADVERTISE_LIMIT = 1000
DEFAULT_INCOMING_PORT = 52780
DEFAULT_LOGGING_RETENTION_DAYS = 90
DEFAULT_PEER_TIMEOUT_SECONDS = 15 * 60  # 15 minutes
DEFAULT_PEER_TIMEOUT_INTERVAL_SECONDS = 10  # 10 seconds
DEFAULT_BOOTSTRAP_RETRY_INTERVAL_SECONDS = 30  # 30 seconds
DEFAULT_STORAGE_INDEX_INTERVAL_SECONDS = 600  # 10 minutes
DEFAULT_INCOMING_QUEUE_SIZE_LIMIT_BYTES = 64 * 1024 * 1024  # 64 MiB
DEFAULT_INCOMING_QUEUE_TIMEOUT_SECONDS = 1.0
DEFAULT_OUTGOING_QUEUE_SIZE_LIMIT_BYTES = 64 * 1024 * 1024  # 64 MiB
DEFAULT_OUTGOING_QUEUE_TIMEOUT_SECONDS = 1.0
DEFAULT_SEED = "bootstrap.astreum.org:52780"
DEFAULT_VERIFICATION_MAX_STALE_SECONDS = 10
DEFAULT_VERIFICATION_MAX_FUTURE_SKEW_SECONDS = 2


def config_setup(config: Dict = {}):
    """
    Normalize configuration values before the node starts.
    """
    chain_str = config.get("chain")
    if chain_str not in {"main", "test"}:
        chain_str = None
    chain_id_raw = config.get("chain_id")
    if chain_id_raw is None:
        chain_id = 1 if chain_str == "main" else 0
    else:
        try:
            chain_id = int(chain_id_raw)
        except (TypeError, ValueError) as exc:
            raise ValueError(
                f"chain_id must be an integer: {chain_id_raw!r}"
            ) from exc
    if chain_str is None:
        chain_str = "main" if chain_id == 1 else "test"
    config["chain"] = chain_str
    config["chain_id"] = chain_id

    hot_limit_raw = config.get(
        "hot_storage_limit", config.get("hot_storage_default_limit", DEFAULT_HOT_STORAGE_LIMIT)
    )
    try:
        config["hot_storage_limit"] = int(hot_limit_raw)
    except (TypeError, ValueError) as exc:
        raise ValueError(
            f"hot_storage_limit must be an integer: {hot_limit_raw!r}"
        ) from exc

    cold_limit_raw = config.get("cold_storage_limit", DEFAULT_COLD_STORAGE_LIMIT)
    try:
        config["cold_storage_limit"] = int(cold_limit_raw)
    except (TypeError, ValueError) as exc:
        raise ValueError(
            f"cold_storage_limit must be an integer: {cold_limit_raw!r}"
        ) from exc

    cold_path_raw = config.get("cold_storage_path")
    if cold_path_raw:
        try:
            path_obj = Path(cold_path_raw)
            path_obj.mkdir(parents=True, exist_ok=True)
            config["cold_storage_path"] = str(path_obj)
        except OSError:
            config["cold_storage_path"] = None
    else:
        config["cold_storage_path"] = None

    advertise_limit_raw = config.get(
        "cold_storage_advertise_limit", DEFAULT_COLD_STORAGE_ADVERTISE_LIMIT
    )
    try:
        advertise_limit = int(advertise_limit_raw)
    except (TypeError, ValueError) as exc:
        raise ValueError(
            "cold_storage_advertise_limit must be an integer: "
            f"{advertise_limit_raw!r}"
        ) from exc
    if advertise_limit < -1:
        raise ValueError(
            "cold_storage_advertise_limit must be -1, 0, or a positive integer"
        )
    config["cold_storage_advertise_limit"] = advertise_limit

    retention_raw = config.get(
        "logging_retention_days",
        config.get("logging_retention", config.get("retention_days", DEFAULT_LOGGING_RETENTION_DAYS)),
    )
    try:
        config["logging_retention_days"] = int(retention_raw)
    except (TypeError, ValueError) as exc:
        raise ValueError(
            f"logging_retention_days must be an integer: {retention_raw!r}"
        ) from exc

    incoming_port_raw = config.get("incoming_port", DEFAULT_INCOMING_PORT)
    try:
        config["incoming_port"] = int(incoming_port_raw)
    except (TypeError, ValueError) as exc:
        raise ValueError(
            f"incoming_port must be an integer: {incoming_port_raw!r}"
        ) from exc

    incoming_queue_limit_raw = config.get(
        "incoming_queue_size_limit", DEFAULT_INCOMING_QUEUE_SIZE_LIMIT_BYTES
    )
    try:
        incoming_queue_limit = int(incoming_queue_limit_raw)
    except (TypeError, ValueError) as exc:
        raise ValueError(
            f"incoming_queue_size_limit must be an integer: {incoming_queue_limit_raw!r}"
        ) from exc
    if incoming_queue_limit < 0:
        raise ValueError("incoming_queue_size_limit must be a non-negative integer")
    config["incoming_queue_size_limit"] = incoming_queue_limit

    incoming_queue_timeout_raw = config.get(
        "incoming_queue_timeout", DEFAULT_INCOMING_QUEUE_TIMEOUT_SECONDS
    )
    try:
        incoming_queue_timeout = float(incoming_queue_timeout_raw)
    except (TypeError, ValueError) as exc:
        raise ValueError(
            f"incoming_queue_timeout must be a number: {incoming_queue_timeout_raw!r}"
        ) from exc
    if incoming_queue_timeout < 0:
        raise ValueError("incoming_queue_timeout must be a non-negative number")
    config["incoming_queue_timeout"] = incoming_queue_timeout

    peer_timeout_raw = config.get("peer_timeout", DEFAULT_PEER_TIMEOUT_SECONDS)
    try:
        peer_timeout = int(peer_timeout_raw)
    except (TypeError, ValueError) as exc:
        raise ValueError(
            f"peer_timeout must be an integer: {peer_timeout_raw!r}"
        ) from exc

    if peer_timeout <= 0:
        raise ValueError("peer_timeout must be a positive integer")

    config["peer_timeout"] = peer_timeout

    interval_raw = config.get("peer_timeout_interval", DEFAULT_PEER_TIMEOUT_INTERVAL_SECONDS)
    try:
        interval = int(interval_raw)
    except (TypeError, ValueError) as exc:
        raise ValueError(
            f"peer_timeout_interval must be an integer: {interval_raw!r}"
        ) from exc

    if interval <= 0:
        raise ValueError("peer_timeout_interval must be a positive integer")

    config["peer_timeout_interval"] = interval

    bootstrap_retry_raw = config.get(
        "bootstrap_retry_interval", DEFAULT_BOOTSTRAP_RETRY_INTERVAL_SECONDS
    )
    try:
        bootstrap_retry_interval = int(bootstrap_retry_raw)
    except (TypeError, ValueError) as exc:
        raise ValueError(
            f"bootstrap_retry_interval must be an integer: {bootstrap_retry_raw!r}"
        ) from exc
    if bootstrap_retry_interval <= 0:
        raise ValueError("bootstrap_retry_interval must be a positive integer")
    config["bootstrap_retry_interval"] = bootstrap_retry_interval

    storage_index_raw = config.get(
        "storage_index_interval", DEFAULT_STORAGE_INDEX_INTERVAL_SECONDS
    )
    try:
        storage_index_interval = int(storage_index_raw)
    except (TypeError, ValueError) as exc:
        raise ValueError(
            f"storage_index_interval must be an integer: {storage_index_raw!r}"
        ) from exc
    if storage_index_interval <= 0:
        raise ValueError("storage_index_interval must be a positive integer")
    config["storage_index_interval"] = storage_index_interval

    outgoing_queue_limit_raw = config.get(
        "outgoing_queue_size_limit", DEFAULT_OUTGOING_QUEUE_SIZE_LIMIT_BYTES
    )
    try:
        outgoing_queue_limit = int(outgoing_queue_limit_raw)
    except (TypeError, ValueError) as exc:
        raise ValueError(
            f"outgoing_queue_size_limit must be an integer: {outgoing_queue_limit_raw!r}"
        ) from exc
    if outgoing_queue_limit < 0:
        raise ValueError("outgoing_queue_size_limit must be a non-negative integer")
    config["outgoing_queue_size_limit"] = outgoing_queue_limit

    outgoing_queue_timeout_raw = config.get(
        "outgoing_queue_timeout", DEFAULT_OUTGOING_QUEUE_TIMEOUT_SECONDS
    )
    try:
        outgoing_queue_timeout = float(outgoing_queue_timeout_raw)
    except (TypeError, ValueError) as exc:
        raise ValueError(
            f"outgoing_queue_timeout must be a number: {outgoing_queue_timeout_raw!r}"
        ) from exc
    if outgoing_queue_timeout < 0:
        raise ValueError("outgoing_queue_timeout must be a non-negative number")
    config["outgoing_queue_timeout"] = outgoing_queue_timeout

    max_stale_raw = config.get(
        "verification_max_stale_seconds", DEFAULT_VERIFICATION_MAX_STALE_SECONDS
    )
    try:
        max_stale = int(max_stale_raw)
    except (TypeError, ValueError) as exc:
        raise ValueError(
            f"verification_max_stale_seconds must be an integer: {max_stale_raw!r}"
        ) from exc
    if max_stale < 0:
        raise ValueError("verification_max_stale_seconds must be a non-negative integer")
    config["verification_max_stale_seconds"] = max_stale

    max_future_raw = config.get(
        "verification_max_future_skew", DEFAULT_VERIFICATION_MAX_FUTURE_SKEW_SECONDS
    )
    try:
        max_future = int(max_future_raw)
    except (TypeError, ValueError) as exc:
        raise ValueError(
            f"verification_max_future_skew must be an integer: {max_future_raw!r}"
        ) from exc
    if max_future < 0:
        raise ValueError("verification_max_future_skew must be a non-negative integer")
    config["verification_max_future_skew"] = max_future

    if "default_seeds" in config:
        raise ValueError("default_seeds is no longer supported; use default_seed")

    if "default_seed" in config:
        default_seed_raw = config["default_seed"]
    else:
        default_seed_raw = DEFAULT_SEED

    if default_seed_raw is None:
        config["default_seed"] = None
    elif isinstance(default_seed_raw, str):
        config["default_seed"] = default_seed_raw
    else:
        raise ValueError("default_seed must be a string or None")

    additional_seeds_raw = config.get("additional_seeds", [])
    if isinstance(additional_seeds_raw, (list, tuple)):
        config["additional_seeds"] = list(additional_seeds_raw)
    else:
        raise ValueError("additional_seeds must be a list of strings")

    return config



from __future__ import annotations

import threading
from queue import Queue
from typing import Any

from astreum.communication.node import connect_node

from .discover import make_discovery_worker
from .worker import make_verify_worker


def verify_blockchain(self: Any):
    """Ensure verification primitives exist, then start discovery and verify workers."""
    connect_node(self)

    self._validation_verify_queue = Queue()

    self.chains = {}
    self.forks = {}
    self.logger.debug(
        "Consensus maps ready for verification (chains=%s, forks=%s)",
        len(self.chains),
        len(self.forks),
    )

    stop_event = getattr(self, "_verify_stop_event", None)
    if stop_event is None:
        stop_event = threading.Event()
        self._verify_stop_event = stop_event
    stop_event.clear()

    discovery_thread = getattr(self, "latest_block_discovery_thread", None)
    if discovery_thread is not None and discovery_thread.is_alive():
        self.logger.debug("Consensus discovery thread already running")
    else:
        discovery_worker = make_discovery_worker(self)
        discovery_thread = threading.Thread(
            target=discovery_worker,
            daemon=True,
            name="latest-block-discovery",
        )
        self.latest_block_discovery_thread = discovery_thread
        discovery_thread.start()
        self.logger.info(
            "Started latest-block discovery thread (%s)", discovery_thread.name
        )

    verify_thread = getattr(self, "verify_thread", None)
    if verify_thread is not None and verify_thread.is_alive():
        self.logger.debug("Consensus verify thread already running")
        return verify_thread

    verify_worker = make_verify_worker(self)
    verify_thread = threading.Thread(
        target=verify_worker, daemon=True, name="verify-worker"
    )
    self.verify_thread = verify_thread
    verify_thread.start()
    self.logger.info("Started verify thread (%s)", verify_thread.name)
    return verify_thread



# Diffie-Hellman key exchange over Curve25519

# x25519.py
from cryptography.hazmat.primitives.asymmetric import x25519
from cryptography.hazmat.primitives.asymmetric.x25519 import X25519PrivateKey, X25519PublicKey
from typing import Tuple

def generate_key_pair() -> Tuple[X25519PrivateKey, X25519PublicKey]:
    """
    Generate an X25519 private and public key pair.
    
    Returns:
        Tuple[X25519PrivateKey, X25519PublicKey]: The generated key pair.
    """
    private_key: X25519PrivateKey = x25519.X25519PrivateKey.generate()
    public_key: X25519PublicKey = private_key.public_key()
    return private_key, public_key

def generate_shared_key(private_key: X25519PrivateKey, peer_public_key: X25519PublicKey) -> bytes:
    """
    Generate a shared key using the provided private key and peer's public key.
    
    Args:
        private_key (X25519PrivateKey): Our private key.
        peer_public_key (X25519PublicKey): The peer's public key.
        
    Returns:
        bytes: The shared key.
    """
    shared_key: bytes = private_key.exchange(peer_public_key)
    return shared_key



from typing import Any, List, Optional, Tuple

from ...storage.models.atom import Atom, AtomKind

ZERO32 = b"\x00" * 32
ERROR_SYMBOL = "error"


class Expr:
    class ListExpr:
        def __init__(self, elements: List['Expr']):
            self.elements = elements
        
        def __repr__(self):
            if not self.elements:
                return "()"
            inner = " ".join(str(e) for e in self.elements)
            return f"({inner})"

        def to_atoms(self):
            return Expr.to_atoms(self)
        
    class Symbol:
        def __init__(self, value: str):
            self.value = value

        def __repr__(self):
            return f"{self.value}"

        def to_atoms(self):
            return Expr.to_atoms(self)
        
    class Bytes:
        def __init__(self, value: bytes):
            self.value = value

        def __repr__(self):
            int_value = int.from_bytes(self.value, "big") if self.value else 0
            return f"{int_value}"

        def to_atoms(self):
            return Expr.to_atoms(self)
    @classmethod
    def from_atoms(cls, node: Any, root_hash: bytes) -> "Expr":
        """Rebuild an expression tree from stored atoms."""
        if not isinstance(root_hash, (bytes, bytearray)):
            raise TypeError("root hash must be bytes-like")

        storage_get = getattr(node, "storage_get", None)
        if not callable(storage_get):
            raise TypeError("node must provide a callable 'storage_get'")

        expr_id = bytes(root_hash)

        def _require(atom_id: Optional[bytes], context: str):
            if not atom_id:
                raise ValueError(f"missing atom id while decoding {context}")
            atom = storage_get(atom_id)
            if atom is None:
                raise ValueError(f"missing atom data while decoding {context}")
            return atom

        def _atom_kind(atom: Any) -> Optional[AtomKind]:
            kind_value = getattr(atom, "kind", None)
            if isinstance(kind_value, AtomKind):
                return kind_value
            if isinstance(kind_value, int):
                try:
                    return AtomKind(kind_value)
                except ValueError:
                    return None
            return None

        type_atom = _require(expr_id, "expression atom")

        kind_enum = _atom_kind(type_atom)
        if kind_enum is None:
            raise ValueError("expression atom missing kind")

        if kind_enum is AtomKind.SYMBOL:
            try:
                return cls.Symbol(type_atom.data.decode("utf-8"))
            except UnicodeDecodeError as exc:
                raise ValueError("symbol atom is not valid utf-8") from exc

        if kind_enum is AtomKind.BYTES:
            return cls.Bytes(type_atom.data)

        if kind_enum is AtomKind.LIST:
            # Empty list sentinel: zero-length payload and no next pointer.
            if len(type_atom.data) == 0 and type_atom.next_id == ZERO32:
                return cls.ListExpr([])

            elements: List[Expr] = []
            current_atom = type_atom
            idx = 0
            while True:
                child_hash = current_atom.data
                if not child_hash:
                    raise ValueError("list element missing child hash")
                if len(child_hash) != len(ZERO32):
                    raise ValueError("list element hash has unexpected length")
                child_expr = cls.from_atoms(node, child_hash)
                elements.append(child_expr)
                next_id = current_atom.next_id
                if next_id == ZERO32:
                    break
                next_atom = _require(next_id, f"list element {idx}")
                next_kind = _atom_kind(next_atom)
                if next_kind is not AtomKind.LIST:
                    raise ValueError("list chain contains non-list atom")
                current_atom = next_atom
                idx += 1
            return cls.ListExpr(elements)

        raise ValueError(f"unknown expression kind: {kind_enum}")

    @staticmethod
    def to_atoms(e: "Expr") -> Tuple[bytes, List[Atom]]:
        def symbol(value: str) -> Tuple[bytes, List[Atom]]:
            atom = Atom(
                data=value.encode("utf-8"),
                kind=AtomKind.SYMBOL,
            )
            return atom.object_id(), [atom]

        def bytes_value(data: bytes) -> Tuple[bytes, List[Atom]]:
            atom = Atom(
                data=data,
                kind=AtomKind.BYTES,
            )
            return atom.object_id(), [atom]

        def lst(items: List["Expr"]) -> Tuple[bytes, List[Atom]]:
            acc: List[Atom] = []
            child_hashes: List[bytes] = []
            for it in items:
                h, atoms = Expr.to_atoms(it)
                acc.extend(atoms)
                child_hashes.append(h)
            next_hash = ZERO32
            elem_atoms: List[Atom] = []
            for h in reversed(child_hashes):
                a = Atom(data=h, next_id=next_hash, kind=AtomKind.LIST)
                next_hash = a.object_id()
                elem_atoms.append(a)
            elem_atoms.reverse()
            if elem_atoms:
                head = elem_atoms[0].object_id()
            else:
                empty_atom = Atom(data=b"", kind=AtomKind.LIST)
                elem_atoms = [empty_atom]
                head = empty_atom.object_id()
            return head, acc + elem_atoms

        if isinstance(e, Expr.Symbol):
            return symbol(e.value)
        if isinstance(e, Expr.Bytes):
            return bytes_value(e.value)
        if isinstance(e, Expr.ListExpr):
            return lst(e.elements)
        raise TypeError("unknown Expr variant")

def _expr_generate_id(expr) -> bytes:
    expr_id, _ = Expr.to_atoms(expr)
    return expr_id


def _expr_cached_id(expr) -> bytes:
    cached = getattr(expr, "_cached_id", None)
    if cached is None:
        cached = _expr_generate_id(expr)
        setattr(expr, "_cached_id", cached)
    return cached


for _expr_cls in (Expr.ListExpr, Expr.Symbol, Expr.Bytes):
    _expr_cls.generate_id = _expr_generate_id  # type: ignore[attr-defined]
    _expr_cls.id = property(_expr_cached_id)  # type: ignore[attr-defined]


def error_expr(topic: str, message: str) -> Expr.ListExpr:
    """Encode an error as (error <topic-bytes> <message-bytes>)."""
    try:
        topic_bytes = topic.encode("utf-8")
    except UnicodeEncodeError as exc:
        raise ValueError("error topic must be valid utf-8") from exc
    try:
        message_bytes = message.encode("utf-8")
    except UnicodeEncodeError as exc:
        raise ValueError("error message must be valid utf-8") from exc
    return Expr.ListExpr([
        Expr.Symbol(ERROR_SYMBOL),
        Expr.Bytes(topic_bytes),
        Expr.Bytes(message_bytes),
    ])

def get_expr_list_from_storage(self, key: bytes) -> Optional["ListExpr"]:
        """Load a list expression from storage using the given atom list root hash."""
        atoms = self.get_atom_list_from_storage(root_hash=key)
        if atoms is None:
            return None
        
        expr_list = []
        for atom in atoms:
            match atom.kind:
                case AtomKind.SYMBOL:
                    expr_list.append(Expr.Symbol(atom.data))
                case AtomKind.BYTES:
                    expr_list.append(Expr.Bytes(atom.data))
                case AtomKind.LIST:
                    expr_list.append(Expr.ListExpr([
                        Expr.Bytes(atom.data),
                        Expr.Symbol("ref")
                    ]))

        expr_list.reverse()
        return Expr.ListExpr(expr_list)



from __future__ import annotations

import time
from datetime import datetime, timedelta, timezone
from typing import TYPE_CHECKING, Any

from ..models.message import Message
from ..util import address_str_to_host_and_port

if TYPE_CHECKING:
    from .. import Node


def _queue_bootstrap_handshakes(node: "Node") -> int:
    outgoing_queue = getattr(node, "outgoing_queue", None)
    relay_public_key = getattr(node, "relay_public_key", None)
    if outgoing_queue is None or relay_public_key is None:
        return 0

    default_seed = node.config.get("default_seed")
    additional_seeds = node.config.get("additional_seeds", [])
    bootstrap_peers = list(additional_seeds)
    if default_seed is not None:
        bootstrap_peers.insert(0, default_seed)
    if not bootstrap_peers:
        return 0

    try:
        incoming_port = int(node.config.get("incoming_port", 0))
        content = incoming_port.to_bytes(2, "big", signed=False)
    except (TypeError, ValueError, OverflowError):
        return 0

    handshake_message = Message(
        handshake=True,
        sender=relay_public_key,
        content=content,
    )
    handshake_bytes = handshake_message.to_bytes()
    sent = 0
    for addr in bootstrap_peers:
        try:
            host, port = address_str_to_host_and_port(addr)
        except Exception as exc:
            node.logger.warning("Invalid bootstrap address %s: %s", addr, exc)
            continue
        outgoing_queue.put((handshake_bytes, (host, port)))
        node.logger.info("Retrying bootstrap handshake to %s:%s", host, port)
        sent += 1
    return sent


def manage_peer(node: "Node") -> None:
    """Continuously evict peers whose timestamps exceed the configured timeout."""
    node.logger.info(
        "Peer manager started (timeout=%3ds, interval=%3ds)",
        node.config["peer_timeout"],
        node.config["peer_timeout_interval"],
    )
    stop = getattr(node, "communication_stop_event", None)
    while stop is None or not stop.is_set():
        timeout_seconds = node.config["peer_timeout"]
        interval_seconds = node.config["peer_timeout_interval"]
        try:
            peers = getattr(node, "peers", None)
            peer_route = getattr(node, "peer_route", None)
            if not isinstance(peers, dict) or peer_route is None:
                time.sleep(interval_seconds)
                continue

            cutoff = datetime.now(timezone.utc) - timedelta(seconds=timeout_seconds)
            stale_keys = []
            with node.peers_lock:
                for peer_key, peer in list(peers.items()):
                    if peer.timestamp < cutoff:
                        stale_keys.append(peer_key)

            removed_count = 0
            for peer_key in stale_keys:
                removed = node.remove_peer(peer_key)
                if removed is None:
                    continue
                removed_count += 1
                try:
                    peer_route.remove_peer(peer_key)
                except Exception:
                    node.logger.debug(
                        "Unable to remove peer %s from route",
                        peer_key.hex(),
                    )
                node.logger.debug(
                    "Evicted stale peer %s last seen at %s",
                    peer_key.hex(),
                    getattr(removed, "timestamp", None),
                )

            if removed_count:
                node.logger.info("Peer manager removed %s stale peer(s)", removed_count)

            try:
                with node.peers_lock:
                    peer_count = len(peers)
            except Exception:
                peer_count = len(getattr(node, "peers", {}) or {})
            if peer_count == 0:
                bootstrap_interval = node.config.get("bootstrap_retry_interval", 0)
                now = time.time()
                last_attempt = getattr(node, "_bootstrap_last_attempt", 0.0)
                if bootstrap_interval and (now - last_attempt) >= bootstrap_interval:
                    sent = _queue_bootstrap_handshakes(node)
                    if sent:
                        node._bootstrap_last_attempt = now
        except Exception:
            node.logger.exception("Peer manager iteration failed")

        if stop is not None and stop.wait(interval_seconds):
            break

    node.logger.info("Peer manager stopped")



from typing import Optional


class Meter:
    def __init__(self, enabled: bool = True, limit: Optional[int] = None):
        self.enabled = enabled
        self.limit: Optional[int] = limit
        self.used: int = 0

    def charge_bytes(self, n: int) -> bool:
        if not self.enabled:
            return True
        if n < 0:
            n = 0
        if self.limit is not None and (self.used + n) >= self.limit:
            return False
        self.used += n
        return True


from typing import List


def tokenize(source: str) -> List[str]:
    tokens: List[str] = []
    cur: List[str] = []
    n = len(source)
    i = 0

    def flush_cur() -> None:
        if cur:
            tokens.append("".join(cur))
            cur.clear()

    def skip_line_comment(idx: int) -> int:
        while idx < n and source[idx] != "\n":
            idx += 1
        return idx

    def skip_ws_and_comments(idx: int) -> int:
        while idx < n:
            ch = source[idx]
            if ch.isspace():
                flush_cur()
                idx += 1
                continue
            if ch == ";":
                flush_cur()
                idx = skip_line_comment(idx + 1)
                continue
            break
        return idx

    def skip_expression(idx: int) -> int:
        idx = skip_ws_and_comments(idx)
        if idx >= n:
            return n
        ch = source[idx]
        if ch == "(":
            depth = 0
            while idx < n:
                ch = source[idx]
                if ch == "(":
                    depth += 1
                    idx += 1
                    continue
                if ch == ")":
                    depth -= 1
                    idx += 1
                    if depth == 0:
                        break
                    continue
                if ch == ";":
                    idx = skip_line_comment(idx + 1)
                    continue
                if ch == "#" and idx + 1 < n and source[idx + 1] == ";":
                    idx = skip_expression(idx + 2)
                    continue
                idx += 1
            return idx
        if ch == ")":
            return idx + 1
        while idx < n:
            ch = source[idx]
            if ch.isspace() or ch in ("(", ")", ";"):
                break
            if ch == "#" and idx + 1 < n and source[idx + 1] == ";":
                break
            idx += 1
        return idx

    while i < n:
        i = skip_ws_and_comments(i)
        if i >= n:
            break
        ch = source[i]
        if ch == "#" and i + 1 < n and source[i + 1] == ";":
            flush_cur()
            i = skip_expression(i + 2)
            continue
        if ch in ("(", ")"):
            flush_cur()
            tokens.append(ch)
            i += 1
            continue
        cur.append(ch)
        i += 1

    flush_cur()
    return tokens
